/*******************************************************************************
* FILE NAME: user_routines_fast.c <FRC VERSION>
*
* DESCRIPTION:
*  This file is where the user can add their custom code within the framework
*  of the routines below. 
*
* USAGE:
*  You can either modify this file to fit your needs, or remove it from your 
*  project and replace it with a modified copy. 
*
* OPTIONS:  Interrupts are disabled and not used by default.
*
*******************************************************************************/

#include <stdio.h>

#include "ifi_aliases.h"
#include "ifi_default.h"
#include "ifi_utilities.h"
#include "user_routines.h"
#include "serial_ports.h"
#include "encoder.h"
#include "camera.h"
#include "tracking.h"
#include "trajectory.h"
#include "iomap.h"
#include "strategy.h"

/*** DEFINE USER VARIABLES AND INITIALIZE THEM HERE ***/
volatile long timer1_ticks = 0;

#ifdef ENABLE_ENCODER_DEBUG
unsigned char lencoder_prev_state = 0;
unsigned char rencoder_prev_state = 0;
#endif

static void shooter_timer(void);
int cur_rps_left = 0;
int cur_rps_right = 0;
int rps_measured = 0;
int autonomous_target_rps = 0;

#ifdef ENABLE_SPEED_CLOCK
static long speed_ticks = 0;
static int speed_clocking = 0;
static void speed_timer(void);
#endif

// variables
typedef enum StartPos 
  { START_NEAR, START_MIDDLE, START_FAR } StartPos;

static char Get_Autonomous_Selection(void);
static int Get_Autonomous_Delay(void);
static StartPos Get_Starting_Position(void);


/*******************************************************************************
* FUNCTION NAME: InterruptVectorLow
* PURPOSE:       Low priority interrupt vector
* CALLED FROM:   nowhere by default
* ARGUMENTS:     none
* RETURNS:       void
* DO NOT MODIFY OR DELETE THIS FUNCTION 
*******************************************************************************/
#pragma code InterruptVectorLow = LOW_INT_VECTOR
void InterruptVectorLow (void)
{
  _asm
    goto InterruptHandlerLow  /*jump to interrupt routine*/
  _endasm
}


/*******************************************************************************
* FUNCTION NAME: InterruptHandlerLow
* PURPOSE:       Low priority interrupt handler
* If you want to use these external low priority interrupts or any of the
* peripheral interrupts then you must enable them in your initialization
* routine.  Innovation First, Inc. will not provide support for using these
* interrupts, so be careful.  There is great potential for glitchy code if good
* interrupt programming practices are not followed.  Especially read p. 28 of
* the "MPLAB(R) C18 C Compiler User's Guide" for information on context saving.
* CALLED FROM:   this file, InterruptVectorLow routine
* ARGUMENTS:     none
* RETURNS:       void
*******************************************************************************/
#pragma code
#pragma interruptlow InterruptHandlerLow save=PROD,section(".tmpdata") /* You may want to save additional symbols. */

void InterruptHandlerLow ()     
{
	unsigned char Port_B;
	unsigned char Port_B_Delta;

	if (PIR1bits.RC1IF && PIE1bits.RC1IE) // rx1 interrupt?
	{
		#ifdef ENABLE_SERIAL_PORT_ONE_RX
		Rx_1_Int_Handler(); // call the rx1 interrupt handler (in serial_ports.c)
		#endif
	}                              
	else if (PIR3bits.RC2IF && PIE3bits.RC2IE) // rx2 interrupt?
	{
		#ifdef ENABLE_SERIAL_PORT_TWO_RX
		Rx_2_Int_Handler(); // call the rx2 interrupt handler (in serial_ports.c)
		#endif
	} 
	else if (PIR1bits.TX1IF && PIE1bits.TX1IE) // tx1 interrupt?
	{
		#ifdef ENABLE_SERIAL_PORT_ONE_TX
		Tx_1_Int_Handler(); // call the tx1 interrupt handler (in serial_ports.c)
		#endif
	}                              
	else if (PIR3bits.TX2IF && PIE3bits.TX2IE) // tx2 interrupt?
	{
		#ifdef ENABLE_SERIAL_PORT_TWO_TX
		Tx_2_Int_Handler(); // call the tx2 interrupt handler (in serial_ports.c)
		#endif
	}
	else if (INTCON3bits.INT2IF && INTCON3bits.INT2IE) // encoder 1 interrupt?
	{ 
		INTCON3bits.INT2IF = 0; // clear the interrupt flag
		#ifdef ENABLE_ENCODER_1
		Encoder_1_Int_Handler(); // call the left encoder interrupt handler (in encoder.c)
		#endif
	}
	else if (INTCON3bits.INT3IF && INTCON3bits.INT3IE) // encoder 2 interrupt?
	{
		INTCON3bits.INT3IF = 0; // clear the interrupt flag
		#ifdef ENABLE_ENCODER_2
		Encoder_2_Int_Handler(); // call right encoder interrupt handler (in encoder.c)
		#endif
	}
	else if (INTCONbits.RBIF && INTCONbits.RBIE) // encoder 3-6 interrupt?
	{
		Port_B = PORTB; // remove the "mismatch condition" by reading port b            
		INTCONbits.RBIF = 0; // clear the interrupt flag
		Port_B_Delta = Port_B ^ Old_Port_B; // determine which bits have changed
		Old_Port_B = Port_B; // save a copy of port b for next time around
	 
		if(Port_B_Delta & 0x10) // did external interrupt 3 change state?
		{
			#ifdef ENABLE_ENCODER_3
			Encoder_3_Int_Handler(Port_B & 0x10 ? 1 : 0); // call the encoder 3 interrupt handler (in encoder.c)
			#endif
		}
		if(Port_B_Delta & 0x20) // did external interrupt 4 change state?
		{
			#ifdef ENABLE_ENCODER_4
			Encoder_4_Int_Handler(Port_B & 0x20 ? 1 : 0); // call the encoder 4 interrupt handler (in encoder.c)
			#endif
		}
		if(Port_B_Delta & 0x40) // did external interrupt 5 change state?
		{
			#ifdef ENABLE_ENCODER_5
			Encoder_5_Int_Handler(Port_B & 0x40 ? 1 : 0); // call the encoder 5 interrupt handler (in encoder.c)
			#endif
		}
		if(Port_B_Delta & 0x80) // did external interrupt 6 change state?
		{
			#ifdef ENABLE_ENCODER_6
			Encoder_6_Int_Handler(Port_B & 0x80 ? 1 : 0); // call the encoder 6 interrupt handler (in encoder.c)
			#endif
		}
  } else if (PIR1bits.TMR1IF && PIE1bits.TMR1IE) {
		PIR1bits.TMR1IF = 0; // resets Timer1 overflow
    // Timer1 counter incremented every 12.5ms
    TMR1L = 0xF6;
    TMR1H = 0xC2;
    ++timer1_ticks; 
  }
}


/*******************************************************************************
* FUNCTION NAME: User_Autonomous_Code
* PURPOSE:       Execute user's code during autonomous robot operation.
* You should modify this routine by adding code which you wish to run in
* autonomous mode.  It will be executed every program loop, and not
* wait for or use any data from the Operator Interface.
* CALLED FROM:   main.c file, main() routine when in Autonomous mode
* ARGUMENTS:     none
* RETURNS:       void
*******************************************************************************/
void User_Autonomous_Code(void)
{
  const char Autonomous_Strategy_Selection = 
                        Get_Autonomous_Selection();
  const StartPos Starting_Position = 
                        Get_Starting_Position();
  const int autofast = Get_Analog_Value(anain_autofast) > 700;
  const int useCam = Get_Analog_Value(anain_box_sw_right) > 700;
  int autodelay = Get_Autonomous_Delay();
  int autoarg, autoarg2;

  /* Initialize all PWMs and Relays when entering Autonomous mode, or else it
     will be stuck with the last values mapped from the joysticks.  Remember, 
     even when Disabled it is reading inputs from the Operator Interface. 
  */
  pwm01 = pwm02 = pwm03 = pwm04 = pwm05 = pwm06 = pwm07 = pwm08 = 127;
  pwm09 = pwm10 = pwm11 = pwm12 = pwm13 = pwm14 = pwm15 = pwm16 = 127;
  relay1_fwd = relay1_rev = relay2_fwd = relay2_rev = 0;
  relay3_fwd = relay3_rev = relay4_fwd = relay4_rev = 0;
  relay5_fwd = relay5_rev = relay6_fwd = relay6_rev = 0;
  relay7_fwd = relay7_rev = relay8_fwd = relay8_rev = 0;

  printf("Starting Pos: %d\r\n", Starting_Position);
  printf("Auto Selection: %d\r\n", Autonomous_Strategy_Selection);
  printf("Autofast: %d\r\n", autofast);
  printf("Autodelay: %d\r\n", autodelay);
  printf("Quirks: %d\r\n", Get_Analog_Value(anain_box_sw_right));
  Reset_Autonomous();
  autonomous_target_rps = 0;

  // Compute any args for the selected autonomous.
  // Do it here so we aren't doing it every loop.
  switch (Autonomous_Strategy_Selection) {
    case 1:
      if (autofast)
        autoarg = 196;
      else
        autoarg = 177;
      break;
    case 2:
      autoarg = DISTANCE_LEN_THIRD;
      if (Starting_Position == START_FAR) {
        autoarg += DISTANCE_CROSS_1SPOS;
        autonomous_target_rps = 30;
      } else if (Starting_Position == START_NEAR) {
        autonomous_target_rps = 30;
      } else {
        autonomous_target_rps = 30;
      }
      break;
    case 3:
      autoarg = DISTANCE_LEN_QUARTER;
      if (Starting_Position == START_FAR)
        autoarg += DISTANCE_CROSS_1SPOS;
      else if (Starting_Position == START_NEAR)
        autoarg -= DISTANCE_CROSS_1SPOS;
      break;
    case 4:
    case 5:
      if (Starting_Position == START_MIDDLE)
        autoarg = DISTANCE_CROSS_1SPOS;
      else if (Starting_Position == START_FAR)
        autoarg = DISTANCE_CROSS_1SPOS * 2;
      else
        autoarg = 0;
      break;
    case 7:
      // Distance, in gear teeth count, to drive out before
      // shooting.
      if (Starting_Position == START_NEAR)
        // About 6.5 feet
        autoarg = 88;
      else if (Starting_Position == START_MIDDLE)
        // About 13 feet
        autoarg = 176;
      else
        // About 19.5 feet
        autoarg = 264;

      autonomous_target_rps = 26;
      if (autofast)
        autoarg2 = 150;
      else
        autoarg2 = 140;
      break;
    default:
      autoarg = 0;
      autoarg2 = 0;
  }
  
  while (autonomous_mode)   /* DO NOT CHANGE! */
  {

	//show off fancy camera angle feedback to user
	//printf("pwm value: %d Cam angle (degrees): %d\r\n", pwm_camera_tilt,Get_Camera_Angle());
	
    if (statusflag.NEW_SPI_DATA)      /* 26.2ms loop area */
    {
      Getdata(&rxdata);   /* DO NOT DELETE, or you will be stuck here forever! */

      Camera_Handler();
      Servo_Track();

      if (autodelay > 0) {
        --autodelay;

      } else {

      	switch(Autonomous_Strategy_Selection)
      	{
          // case 0 is default - do nothing.
      		case 1:
      			Straight_Defense(autoarg, !useCam);
      			break;
      		case 2:
            if (Starting_Position == START_NEAR)
              // Just shoot in place
              Auto_Launch_Handler(1);
            else
              Out_And_Turn(DISTANCE_CROSS_THIRD, 
                           autoarg, 1);
      			break;
      		case 3:
      			Out_And_Turn(DISTANCE_CROSS_THIRD,
                         autoarg, 0);
      			break;
      		case 4:
      			Sideways_Defense(DISTANCE_CROSS_QUARTER, 
                             autoarg);
      			break;
      		case 5:
      			Out_And_Turn(DISTANCE_CROSS_QUARTER, 
                         autoarg, 0);
      			break;
      		case 6:
      			Cross_Defense();
      			break;
      		case 7:
            Drive_And_Shoot(autoarg, autoarg2, useCam);
            break;
      		default:
      			pwm_right_drive = pwm_left_drive = 127;
            break;
      	}
      }

      //Robot_Orient();				
      Putdata(&txdata);   /* DO NOT DELETE, or you will get no PWM outputs! */
    }
    Process_Data_From_Local_IO();

    // Fast-loop debug statements
    //printf("The angle is %d\t The PWM value is %d \r\n",
    //                (int) Get_Camera_Angle(), pwm_camera_tilt);
  }
}

/*******************************************************************************
* FUNCTION NAME: Process_Data_From_Local_IO
* PURPOSE:       Execute user's realtime code.
* You should modify this routine by adding code which you wish to run fast.
* It will be executed every program loop, and not wait for fresh data 
* from the Operator Interface.
* CALLED FROM:   main.c
* ARGUMENTS:     none
* RETURNS:       void
******************************************************************g*************/
void Process_Data_From_Local_IO(void)
{
  /* Add code here that you want to be executed every program loop. */

#ifdef ENABLE_ENCODER_DEBUG
	if (digin_left_wheel_enc != lencoder_prev_state) {
		if (digin_left_wheel_enc) {
			lencoder_prev_state = 1;
		} else {
			lencoder_prev_state = 0;
		}
		printf("LEFT Ticks: %ld\r\n", Get_Left_Encoder_Count());
	}
	if (digin_right_wheel_enc != rencoder_prev_state) {
		if (digin_right_wheel_enc) {
			rencoder_prev_state = 1;
		} else {
			rencoder_prev_state = 0;
		}
		printf("RIGHT Ticks: %ld\r\n", Get_Right_Encoder_Count());
	}
#endif
  

  //renamed from encoder timer-this is for
  //shooter wheels instead
	shooter_timer();

  // This uses the same timer as RPS_SHOOTER.... so
  // we can't use it at the same time....
#ifdef ENABLE_SPEED_CLOCK
	speed_timer();
#endif



	// Watch speed banner sensor

}


static void shooter_timer(void) {
	long ticks = Read_Timer1();
	//printf("Timer Ticks: %ld\r\n", ticks);
	if (ticks > 9) {
    //long x = Get_Encoder_5_Count();
    //if (x < 29 || x > 30)
  	//	printf("Encoder ticks/sec: %ld\r\n", Get_Encoder_5_Count());
    printf("Right RPS: %ld\r\n", Get_Encoder_6_Count());
    printf("Left RPS: %ld\r\n", Get_Encoder_5_Count());
    
    cur_rps_left = Get_Encoder_5_Count();
    cur_rps_right = Get_Encoder_6_Count();
    rps_measured = 1;

    Stop_Timer1();
#ifdef ENABLE_SPEED_CLOCK  
    if (speed_clocking)
      speed_ticks += ticks;
#endif
		Reset_Encoder_5_Count();
    Reset_Encoder_6_Count();
		Start_Timer1();
	}
}






// A function that returns the autonomous selection based upon
// binary readins from selection_bit1, bit2 and bit3
// if an error, returns -1
static char Get_Autonomous_Selection(void)
{
  // These analog ins are active high!
  // Set the compare bar fairly high to ensure a floating 
  // input (resulting from disconnected box) to read out
  // as a '1' on the bit. After the '7 - selection' 
  // adjustment, that'll be strategy 0 (do nothing == safe!)
  int b0 = Get_Analog_Value(anain_autoselect_lsb) < 500 ? 0 : 0x1;
  int b1 = Get_Analog_Value(anain_autoselect_midb) < 500 ? 0 : 0x2;
  int b2 = Get_Analog_Value(anain_autoselect_msb) < 500 ? 0 : 0x4;

  int selection = b2 | b1 | b0;
  printf("AutoSelect Raw Bits: %c%c%c Select: %d\n",
          b2 ? '1' : '0', b1 ? '1' : '0', b0 ? '1' : '0',
          selection);

  return (char)selection;
}

// A function that returns the starting position on the field,
// 3 being farthest from the drivers' station, 1 being closest.
static StartPos Get_Starting_Position(void)
{
  int position;

  int distant = Get_Analog_Value(anain_position_far);
  int close = Get_Analog_Value(anain_position_near);

  // The above-used switches are active low!
  if (close < 20)
  	position = START_NEAR;
  else if (distant < 20)
	  position = START_FAR;
  else
    position = START_MIDDLE;

  return position;
}

static int Get_Autonomous_Delay(void)
{
  int en = Get_Analog_Value(anain_autodelay) > 700;
  int ret;

  if (en) {
    // Get pot value, which is 0 -> 1023, but backwords.
    // Reverse it
    ret = 1023 - Get_Analog_Value(anain_box_pot);
    // Divide by 5 to get a delay count of 0 -> 204
    // When used in the main autonomous handler as a 
    // delay counter of the (approx) 26.2 ms slow loops,
    // this yields a delay between 0 -> 5.3 seconds
    ret /= 5;
  } else {
    ret = 0;
  }

  return ret;
}








#ifdef ENABLE_SPEED_CLOCK
#include <math.h>
static void speed_timer(void) {
  static int speed_banner_prev_on = 0;

	if (!digin_speed_banner) {
		if ( speed_banner_prev_on) {
      speed_clocking = 1;
		  printf("Timer started\r\n");
    }
  } else {
    if (!speed_banner_prev_on) {
      int f1;
      f1 = 1778 * 80 / (speed_ticks + timer1_ticks);
      
      printf("speed sensor on for %ld ticks. (%u ms)\r\n", speed_ticks + timer1_ticks, (unsigned)((speed_ticks + timer1_ticks) * 1000 / 80));
      printf("Ball Speed: %d mm/s\r\n", f1);
      speed_ticks = 0;
      speed_clocking = 0;
    }
  }

	speed_banner_prev_on = digin_speed_banner;
}

#endif // ENABLE_SPEED_CLOCK



/*******************************************************************************
*
*	FUNCTION:		Initialize_Timer_1()
*
*	PURPOSE:		Initializes the timer 1 hardware.
*
*	CALLED FROM:	user_routines.c/User_Initialization()
*
*	PARAMETERS:		None
*
*	RETURNS:		Nothing
*
*	COMMENTS:		Place "#include "interrupts.h" in the includes section
*					of user_routines.c then call Initialize_Timer_1() in
*					user_routines.c/User_Initialization().
*
*					Timer 1 documentation starts on page 135 of the data sheet.
*
*******************************************************************************/
void Initialize_Timer_1(void)  
{
  // Timer1 counter incremented every 12.5ms
  TMR1L = 0xF6;
  TMR1H = 0xC2;
							//
	T1CONbits.T1CKPS0 = 1;	// T1CSP1 T1CSP0
	T1CONbits.T1CKPS1 = 1;	//   0      0		1:1 prescaler (clock=10MHz/each tick=100ns)
							//   0      1		1:2 prescaler (clock=5MHz/each tick=200ns)
							//   1      0		1:4 prescaler (clock=2.5MHz/each tick=400ns)
							//   1      1		1:8 prescaler (clock=1.25MHz/each tick=800ns)
							//
	T1CONbits.T1OSCEN = 0;	// 0: timer 1 oscillator disabled (leave at 0 to allow the use of an external clock)
							// 1: timer 1 oscillator enabled (can't be used because of hardware constraints)
							//
	T1CONbits.TMR1CS = 0;	// 0: use the internal clock
							// 1: use an external clock on RC0/T1OSO/T13CLK (rc_dig_in14 on full-size controller)
							//
	T1CONbits.RD16 = 1;		// 0: timer 1 register operations are done in two 8-bit accesses
							// 1: timer 1 register operations are done in one 16-bit access
							//    In this mode, reading TMR1L will latch a copy of TMR1H into a buffer
							//    mapped to the TMR1H memory address. Conversely, a write to the buffer
							//    followed by a write to the TMR1L register will update the entire 16-bit
							//    timer at once. This solves the problem where the timer may overflow
							//    between two 8-bit accesses. Here's an example of how to do a 16-bit read:
							//
							//    unsigned char Temp_Buf; // 8-bit temporary buffer
							//    unsigned int Timer_Snapshot; // 16-bit variable
							//
							//    Temp_Buf = TMR1L; // TMR1L must be read before TMR1H
							//    Timer_Snapshot = TMR1H;
		 					//    Timer_Snapshot <<= 8; // move TMR1H data to the upper half of the variable
							//    Timer_Snapshot += Temp_Buf; // we now have all sixteen bits  
							// 
	IPR1bits.TMR1IP = 0;	// 0: timer 1 overflow interrupt is low priority (leave at 0 on IFI controllers)
							// 1: timer 1 overflow interrupt is high priority
							//
	PIR1bits.TMR1IF = 0;	// 0: timer 1 overflow hasn't happened (set to 0 before enabling the interrupt)
							// 1: timer 1 overflow has happened
							//

  timer1_ticks =0;

	PIE1bits.TMR1IE = 1;	// 0: disable timer 1 interrupt on overflow (i.e., a transition from FFFF->0)
							// 1: enable timer 1 interrupt on overflow (i.e., a transition from FFFF->0)
							//	
	T1CONbits.TMR1ON = 1;	// 0: timer 1 is disabled
							// 1: timer 1 is enabled (running)
}


void Start_Timer1(void)
{
  // Timer1 counter incremented every 12.5ms
	TMR1L = 0xF6;
  TMR1H = 0xC2;
	PIE1bits.TMR1IE = 0;
  timer1_ticks = 0;
	PIE1bits.TMR1IE = 1;

	T1CONbits.TMR1ON = 1;
}

void Stop_Timer1(void)
{
	T1CONbits.TMR1ON = 0;
}

long Read_Timer1(void)
{
  long ret;
	PIE1bits.TMR1IE = 0;
  ret = timer1_ticks;
	PIE1bits.TMR1IE = 1;
  return ret;
}


/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
